/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Damien Fournier
 */

package org.objectweb.fractal.juliac;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DependencyResolutionRequiredException;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.metadata.ArtifactMetadataRetrievalException;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.metadata.ResolutionGroup;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.juliac.api.JuliacModuleItf;
import org.objectweb.fractal.juliac.api.RuntimeClassNotFoundException;
import org.objectweb.fractal.juliac.conf.Constants;
import org.objectweb.fractal.juliac.conf.GeneratorMap;
import org.objectweb.fractal.juliac.conf.InterceptorSourceCodeGeneratorMap;
import org.objectweb.fractal.juliac.conf.JDKLevel;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.conf.Module;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;

/**
 * Abstract root class for the Juliac Maven plugin.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Damien Fournier <Damien.Fournier@inria.fr>
 */
public abstract class AbstractJuliacMojo extends AbstractMojo
implements JuliacMojoItf {
    
    // ---------------------------------------------------------------------
    // Plugin configuration parameters
    // ---------------------------------------------------------------------
    
    /**
     * System arguments.
     * 
     * @parameter
     */
    private Map<String,String> sysargs;
    
    /**
     * The list of source roots.
     * 
     * @parameter
     */
    private List<String> srcs;
    
    /**
     * The source code compatibility level.
     * Default value: 1.5.
     * 
     * @parameter
     */
    private String source;
    
    /**
     * The bytecode compatibility level.
     * Default value: 1.5.
     * 
     * @parameter
     */
    private String target;
    
    /**
     * The source artifact descriptor which contain the mixin layers. An
     * artifact descriptor is a colon separated string of the form
     * <code>groupId:artifactId:version:sources</code>.
     * 
     * @parameter
     */
    private List<String> mixins;
    
    /**
     * The root package for generated code.
     * Default value: none.
     * 
     * @parameter
     */
    private String pkgRoot;
    
    /**
     * The jar file containing the compiled source code.
     * 
     * @parameter
     */
    private String genclassjar;
    
    /**
     * The interceptor class generators.
     * 
     * @parameter
     * @since 2.2.6
     */
    private Map<String,String> iclassgs;
    
    /**
     * The interceptor source code generators.
     * 
     * @parameter
     */
    private Map<String,String> icodegs;
    
    /**
     * The ADLs which are to be compiled with Juliac.
     * 
     * @parameter
     */
    private List<String> adls;
    
    /**
     * The membranes which are to be compiled with Juliac.
     * 
     * @parameter
     */
    private List<String> membranes;
    
    /**
     * The components which are to be compiled with Juliac.
     * 
     * @parameter
     */
    private List<Component> components;
    
    /**
     * The proxies which are to be compiled with Juliac.
     * 
     * @parameter
     * @since 2.3.1
     */
    private List<Proxy> proxies;
    
    /**
     * Juliac artifact descriptors. An artifact descriptor is a colon separated
     * string of the form <code>groupId:artifactId:version</code>. Artifact
     * descriptors returned by {@link Module#getArtifactDescForDefaultModules()}
     * are taken as default if no artifact descriptor is specified.
     * 
     * @parameter
     */
    private List<String> artifacts;
        
    /**
     * The fully qualified names of the modules to be loaded. Module classes
     * must implement {@link JuliacModuleItf}. The constant values defined in
     * {@link Module} can be used in replacement for module implementation class
     * names.
     * 
     * @parameter
     * @since 2.4
     */
    private List<String> modules;

    /**
     * <code>false</code> to prevent compilation of input code.
     * Default value: <code>true</code>.
     * 
     * @parameter
     */
    private boolean compileInput = true;
    
    /**
     * <code>false</code> to prevent compilation of generated code.
     * Default value: <code>true</code>.
     * 
     * @parameter
     */
    private boolean compileGenerated = true;
    
    /**
     * <code>true</code> to obtain debug information about the code which is
     * being generated. Default value: <code>false</code>.
     * 
     * @parameter
     */
    private boolean debug = false;

    /**
     * <code>true</code> to obtain warnings about the code which is being
     * compiled. Default value: <code>false</code>.
     * 
     * @parameter
     */
    private boolean compilationWarnings = false;
    
    /**
     * <code>true</code> to exit the plugin without throwing an exception.
     * Default value: <code>false</code>.
     * 
     * When set to <code>true</code>, the idea is to report potential errors but
     * to exit normally. This feature originates from a discussion with Nicolas
     * Salatge and Julien Lesbegueries from Petals Link, who want to be able to
     * run commands like <code>mvn eclipse:eclipse</code> even if this plugin
     * reports an error. The issue is not specific to this plugin but to the
     * fact that, in most cases, this plugin is associated with the
     * <code>generate-sources</code> lifecycle phase which is run when commands
     * like <code>mvn eclipse:eclipse</code> are invoked.
     * 
     * @parameter
     * @since 2.2.6.1
     */
    private boolean gracefulExit = false;

    
    // ---------------------------------------------------------------------
    // Maven project data structures
    // ---------------------------------------------------------------------
    
    /** @component */
    private ArtifactFactory artifactFactory;
    
    /** @component */
    protected ArtifactMetadataSource artifactMetadataSource;
    
    /** @component */
    private ArtifactResolver resolver;

    /** @parameter expression="${localRepository}" */
    private ArtifactRepository localRepository;

    /** @parameter expression="${project}" */
    protected MavenProject project;
    
    /** @parameter expression="${project.remoteArtifactRepositories}" */
    private List<ArtifactRepository> remoteRepositories;
    
    /** @parameter expression="${plugin.artifacts}" */
    private List<Artifact> pluginArtifacts;
        
    
    // ---------------------------------------------------------------------
    // Abstract method
    // ---------------------------------------------------------------------
    
    /**
     * Return the directory containing the generated source code.
     * 
     * @since 2.3
     */
    abstract protected String getGenSrc();
    
    /**
     * Return the directory containing the compiled source code.
     * 
     * @since 2.3
     */
    abstract protected String getGenClass();
    
    /**
     * Add resources to the specified class path elements.
     * 
     * @since 2.3
     */
    abstract protected void addResources( JuliacMojoClassPathElements cpe );
    
    /**
     * Add dependencies to the specified class path elements.
     * 
     * @since 2.3
     */
    abstract protected void addPaths( JuliacMojoClassPathElements cpe )
    throws DependencyResolutionRequiredException;
    
    /**
     * Add the specified path to the source root.
     * 
     * @since 2.3
     */
    abstract protected void addSourceRoot( String genDirName, Logger logger );
    
    
    // ---------------------------------------------------------------------
    // Plugin entry point
    // ---------------------------------------------------------------------
    
    public void execute() throws MojoExecutionException {
        
        try {
            executeWithExceptions();
        }
        catch( Throwable t ) {        	
			if( ! gracefulExit ) {
                throw new MojoExecutionException("",t);
        	}
        }
    }

    
    // --------------------------------------------------------------------
    // Method for invoking Juliac
    // --------------------------------------------------------------------

    private void executeWithExceptions()
    throws
        MojoExecutionException, IOException,
        InstantiationException, MalformedURLException,
        DependencyResolutionRequiredException, ArtifactResolutionException,
        ArtifactNotFoundException, RuntimeClassNotFoundException,
        ArtifactMetadataRetrievalException {

        Juliac jc = new Juliac();
        JuliacConfig jconf = new JuliacConfig(jc);
        jc.setJuliacConfig(jconf);
        
        File baseDir = project.getBasedir();
        jconf.setBaseDir(baseDir);

        /*
         * Set the log levels.
         */
        Logger logger = jconf.getLogger();        
        if(debug) {
        	JuliacConfig.setLevel(logger,Level.FINE);
        }
        jconf.setCompilationWarnings(compilationWarnings);

        /*
         * System arguments. <sysargs>
         */
        Map<String,String> oldsysargs = new HashMap<String,String>();
        if( sysargs != null ) {
            for (Map.Entry<String,String> sysarg : sysargs.entrySet()) {
                
                String name = sysarg.getKey();
                String value = sysarg.getValue();
                value = Utils.trimHeadingAndTrailingBlanks(value);
                
                /*
                 * Save the original value of the property in orig (for
                 * re-installation at the end of the method), and set the new
                 * value.
                 */
                String orig = System.getProperty(name);
                oldsysargs.put(name,orig);
                System.setProperty(name,value);
                
                logger.fine("System property: "+name+'='+value);
            }
        }
        
        /*
         * Source root. <srcs>
         * 
         * Transmitting the base directory is needed when the value for <src> is
         * a relative path and that the base directory is not the directory from
         * which mvn is launched (e.g. if mvn is launched recursively from a
         * directory located higher in the hierarchy.)
         */
        JuliacMojoClassPathElements cpe = new JuliacMojoClassPathElements(this);        
        String[] locations = trimHeadingAndTrailingBlanks(this.srcs);
        String[] srcs =
            locations.length == 0 ?
            new String[0] :
            cpe.addLocationDescriptors(baseDir,locations);
        jconf.addSrcs(srcs);
        
        /*
         * Add mixins <mixins> to the classpath.
         * 
         * See aformentioned comment about baseDir.
         */
        locations = trimHeadingAndTrailingBlanks(mixins);
        String[] srclibs =
            locations.length == 0 ?
            new String[0] :
            cpe.addLocationDescriptors(baseDir,locations);
        jconf.addSrclibs(srclibs);
        
        /*
         * The root package.
         */
        if( pkgRoot == null ) {
            // Retrieve the default value
            pkgRoot = jconf.getPkgRoot();
        }
        else {
            jconf.setPkgRoot(pkgRoot);
        }
        
        /*
         * The directory containing the generated source code. <gensrc>
         */
        String gensrc = getGenSrc();
        jconf.setGenDirName(gensrc);
        
        /*
         * The directory containing the compiled source code. <genclass>
         */
        String genclass = getGenClass();
        jconf.setClassDirName(genclass);
        
        /*
         * The jar file name containing the compiled source code. <genclassjar>
         */
        String genclassjar =
        	Utils.trimHeadingAndTrailingBlanks(this.genclassjar);
        if( genclassjar != null ) {
        	if( ! compileGenerated ) {
        		String msg =
    				"<genclassjar> can not be used if <compiledGenerated> is "+
    				"set to false";
        		logger.severe(msg);
        		return;
        	}
        }
        
        /*
         * Even though the default value has been retrieved with
         * getClassDirName(), invoking setClassDirName() has the effect of
         * creating the genclass directory (by default target/classes.)
         * 
         * This creation is of particular importance for the URLClassLoader
         * instantiated below and which relies on URLs created via
         * File.toURI().toURL(). If the directory does not exist, the
         * corresponding URL does not contain a trailing slash character. This
         * omission causes the denoted path to be interpreted, at least under
         * Windows, as a file and not as a directory. 
         */
        jconf.setClassDirName(genclass);
        
        /*
         * The source code compatibility level.
         */
        if( source == null ) {
            // Retrieve the default value
            source = jconf.getSourceLevel().getVersion();
        }
        else {
            String source = Utils.trimHeadingAndTrailingBlanks(this.source);
            if( source.length() != 0 ) {
            	JDKLevel level = JDKLevel.valueOfIAE(source);
                jconf.setSourceLevel(level);
            }
        }
        
        /*
         * The bytecode compatibility level.
         */
        if( target == null ) {
            // Retrieve the default value
            target = jconf.getTargetLevel().getVersion();
        }
        else {
            String target = Utils.trimHeadingAndTrailingBlanks(this.target);
            if( target.length() != 0 ) {
            	JDKLevel level = JDKLevel.valueOfIAE(target);
                jconf.setTargetLevel(level);
            }
        }
        
        /*
         * Interceptor class generators <icodegs>.
         */
    	GeneratorMap<InterceptorClassGenerator> iclgs =
    		jconf.interceptorClassGenerators();
        if( iclassgs == null ) {
            iclassgs = new HashMap<String,String>();
        }
        for (Map.Entry<String,String> entry : iclassgs.entrySet()) {
            String asmcg = entry.getKey();
            String scg = entry.getValue();
            scg = Utils.trimHeadingAndTrailingBlanks(scg);
            iclgs.set(asmcg,scg);
        }
        
        /*
         * Interceptor source code generators <icodegs>.
         */
        InterceptorSourceCodeGeneratorMap iscgs =
    		jconf.interceptorSourceCodeGenerators();
        if( icodegs == null ) {
            icodegs = new HashMap<String,String>();
        }
        for (Map.Entry<String,String> entry : icodegs.entrySet()) {
            String asmcg = entry.getKey();
            String scg = entry.getValue();
            scg = Utils.trimHeadingAndTrailingBlanks(scg);
        	iscgs.set(asmcg,scg);
        }
        
        /*
         * Juliac modules <modules>.
         */
        List<String> modules = new ArrayList<String>();
        if( this.modules != null && this.modules.size() != 0 ) {
        	for (String module : this.modules) {
                String s = Utils.trimHeadingAndTrailingBlanks(module);
                if( s.length() != 0 ) {
                	modules.add(s);
                }
			}
        }

        /*
         * Juliac module artifacts <artifacts>.
         * Add artifacts corresponding to modules specified with a constant.
         */
        List<String> artifacts = new ArrayList<String>();
        if( this.artifacts != null && this.artifacts.size() != 0 ) {
        	artifacts.addAll(this.artifacts);
        }
    	for (String module : modules) {
    		try {
        		Module m = Module.valueOf(module);
        		String artifact = m.getArtifactDesc();
        		artifacts.add(artifact);
    		}
    		catch( IllegalArgumentException iae ) {
    			// Not a module associated with a constant defined in Module
    		}
		}
        String[] as = trimHeadingAndTrailingBlanks(artifacts);
        cpe.addLocationDescriptors(baseDir,as);
        
        /*
         * Add Maven resources directories to the classpath.
         */
        addResources(cpe);
        
        /*
         * Add dependencies specified in the project to the classpath.
         */
        addPaths(cpe);
        
        /*
         * Add the directory specified in <genclass> to the classpath.
         */
        File classDir = Utils.getFile(baseDir,genclass);
        String classDirName = classDir.getAbsolutePath();
        cpe.addPath(classDirName);
        
        /*
         * Add dependencies specified locally when using this plugin to the
         * classpath.
         * 
         * Don't use project.getPluginArtifacts() which returns a list which
         * contains only the artifact of the plugin itself and not the artifacts
         * which are specified in the <plugin> <dependency> local section.
         * 
         * Moreover retrieving dependencies from the <plugin> <dependency> local
         * section must not be used. This works when the MOJO is used in only
         * one subproject, but not when it is used in several subprojects with
         * different configurations. This is the case for example with the
         * julia-mixed/comp and julia-mixed/oo modules where one would want to
         * specify a dependency to juliac-opt-comp in the former case and
         * juliac-opt-oo in the latter. The second configuration is ignored and
         * leads to a class not found exception. It seems that the <plugin>
         * <dependency> local section is not re-read when it has been read once.
         * This is implemented in the maven-core module,
         * org.apache.maven.plugin.DefaultPluginManager#ensurePluginContainerIsComplete
         * method (line 619 for version 2.0.7 of the module) where the first
         * statement of the method:
         *      if( pluginDescriptor.getArtifacts() != null &&
         *          pluginDescriptor.getArtifacts().size() == 1 )
         * skips the resolution of artifacts specified in the <plugin>
         * <dependency> local section when they have been resolved once (this is
         * the case since pluginDescriptor is retained between two executions of
         * the plugin in the same Maven run.)
         * 
         * Although retrieving dependencies from the <plugin> <dependency> local
         * section suffers from the previously mentioned bug, we keep it, just
         * in case one developer would use it.
         */
        cpe.addArtifacts(pluginArtifacts);
        
        /*
         * Define the Juliac class loader.
         */
        URL[] urls = cpe.getURLs();
        if( urls.length != 0 ) {
            
            for (URL url : urls) {
            	logger.fine("Juliac classpath: "+url.toString());
            }
            
            /*
             * Specify getClass().getClassLoader() as the parent class loader.
             * Else classes used by Juliac and the ones dynamically loaded by
             * urlcl do not have a common parent ancestor and can not be casted.
             */
            ClassLoader parent = getClass().getClassLoader();
            ClassLoader urlcl = new URLClassLoader( urls, parent );
            jconf.setClassLoader(urlcl);            
        }

        
        // --------------------------------------------------------------------
        // Configure the Maven compilation plugin
        // --------------------------------------------------------------------

        /*
         * Instruct Maven to consider the location specified in gensrc as a
         * directory where source code is present and may need to be compiled.
         * 
         * Until Juliac 2.2.6 included, we were also adding locations specified
         * in srcs. However this was causing a bug when the Juliac plugin was
         * used jointly with the Clover plugin. The problem was reported by
         * Nicolas Salatge from Petals Link.
         */
        File genDir = Utils.getFile(baseDir,gensrc);
        String genDirName = genDir.getAbsolutePath();
        addSourceRoot(genDirName,logger);
        

        // --------------------------------------------------------------------
        // Invoke Juliac
        // --------------------------------------------------------------------
        
        /*
         * Load modules. User specified modules defined in the artifact being
         * built by this MOJO may need to be compiled. Conversely, a compiler
         * module needs to be loaded for compiling input code. 
         */
        List<String> userDefinedModules = new ArrayList<String>();
        for (String module : modules) {
        	try {
    			jconf.loadModule(module);
        	}
        	catch( RuntimeClassNotFoundException cnfe ) {
        		userDefinedModules.add(module);
        	}
		}
        
        /*
         * Compile input code if requested.
         */
        if( compileInput ) {
            jc.compile();
        }
        
        /*
         * Load user specified modules.
         */
        for (String module : userDefinedModules) {
			jconf.loadModule(module);
		}
        
        /*
         * Code generation for membranes.
         */
        String[] membranes = trimHeadingAndTrailingBlanks(this.membranes);
        for (String membrane : membranes) {
        	logger.info(membrane+"...");
            jc.generate(membrane);
        }
        
        /*
         * Code generation for ADLs.
         */
        String[] adls = trimHeadingAndTrailingBlanks(this.adls);
        for (String adl : adls) {
        	logger.info(adl+"...");
            jc.generate(adl,adl);
        }
        
        /*
         * Code generation for components.
         */
        if( components != null ) {
            for (Component component : components) {
            	logger.info(component+"...");
                Type type = component.toFractalComponentType();
                Object controllerDesc = component.getControllerDesc();
                Object contentDesc = component.getContentDesc();
                jc.generate(type,controllerDesc,contentDesc);
            }
        }
        
        /*
         * Code generation for proxies.
         */
        if( proxies != null ) {
            for (Proxy proxy : proxies) {
            	logger.info(proxy+"...");
                String signature = proxy.getSignature();
                Object controllerDesc = proxy.getControllerDesc();
                jc.generateProxiesImpl(signature,controllerDesc);
            }
        }
        
        /*
         * Compile generated code if requested.
         */
        if( compileGenerated ) {
        	jconf.setCompilationWarnings(false);
            jc.compile();
        	jconf.setCompilationWarnings(compilationWarnings);
        }

        /*
         * Reset the system properties.
         */
        for (Map.Entry<String,String> entry : oldsysargs.entrySet()) {
            String name = entry.getKey();
            String value = entry.getValue();
            if( value == null ) {
                System.clearProperty(name);
            }
            else {
                System.setProperty(name,value);
            }
        }
        
        /*
         * Close Juliac.
         */
        jc.close();
    }


    // --------------------------------------------------------------------
    // Implementation of JuliacMojoItf
    // --------------------------------------------------------------------
    
    /*
     * Various ways for retrieving the version number of the current executing
     * maven-juliac-plugin have been tried, but this method is the only one
     * which proved to be successful.
     * 
     *      @parameter expression="${plugin}"
     *      private PluginDescriptor pd;
     * 
     * seemed to work at first sight, but then
     * DependencyResolutionRequiredException was thrown when trying to access
     * pluginArtifacts.
     */
    public String getPluginVersion() throws IllegalArgumentException {
        
        for (Artifact pa : pluginArtifacts) {
            if( pa.getArtifactId().equals(Constants.ARTIFACT_CORE) ) {
                return pa.getBaseVersion();
            }
        }
        String msg = "Unable to find version for current plugin";
        throw new IllegalArgumentException(msg);
    }
    
    public Artifact resolveArtifact(
        String groupId, String artifactId, String version, String type,
        String classifier )
    throws ArtifactResolutionException, ArtifactNotFoundException {
        
        Artifact artifact =
            artifactFactory.createArtifactWithClassifier(
                groupId, artifactId, version, type, classifier );
        resolveArtifact(artifact);
        return artifact;
    }

    public void resolveArtifact( Artifact artifact )
    throws ArtifactResolutionException, ArtifactNotFoundException {        
        resolver.resolve( artifact, remoteRepositories, localRepository );
    }

    /**
     * Return the list of artifacts depending transitively from the specified
     * ones.
     */
    public List<Artifact> resolveArtifactTransitively( List<Artifact> artifacts )
    throws
        ArtifactResolutionException, ArtifactMetadataRetrievalException,
        ArtifactNotFoundException {

        List<Artifact> ret = new ArrayList<Artifact>();
        for (Artifact artifact : artifacts) {
            List<Artifact> list = resolveArtifactTransitively(artifact);
            for (Artifact l : list) {
                addIfNew(ret,l);
            }
        }
        return ret;
    }
    
    /**
     * Add the specified artifact to the list if the list does not already
     * contain this artifact or if this is a newer version of the artifact. In
     * this case, remove the older version.
     */
    private void addIfNew( List<Artifact> artifacts, Artifact l ) {
        
        String lgid = l.getGroupId();
        String laid = l.getArtifactId();
        String lcla = l.getClassifier();
        
        Artifact toBeRemoved = null;
        boolean toBeAdded = true;
        for (Artifact artifact : artifacts) {
            String groupId = artifact.getGroupId();
            String artifactId = artifact.getArtifactId();
            String classifier = artifact.getClassifier();
            
            if( groupId.equals(lgid) && artifactId.equals(laid) &&
                (classifier==null || classifier.equals(lcla)) ) {

                /*
                 * Artifact#compareTo (see class DefaultArtifact) makes a
                 * comparison on the stringified version number (as returned by
                 * Artifact#getVersion()). Should this be sufficient?
                 */
                String version = artifact.getVersion();
                String lv = l.getVersion();
                int v = version.compareTo(lv);
                
                if( v < 0 ) {
                    /*
                     * artifact is older than l. artifact needs to be removed in
                     * order to prevent conflicts with l which will be added.
                     */
                    toBeRemoved = artifact;
                }
                else {
                    /*
                     * artifact is equals or newer than l. l must not be added.
                     */
                    toBeAdded = false;
                }
                // In all cases, we can skip the rest of the list
                break;
            }
        }
        
        /*
         * Don't remove while iterating on artifacts to avoid
         * ConcurrentModificationException.
         */
        if( toBeRemoved != null ) {
            artifacts.remove(toBeRemoved);
        }
        if(toBeAdded) {
            artifacts.add(l);
        }
    }
    
    /**
     * Return the list of artifacts depending transitively from the specified
     * one.
     */
    private List<Artifact> resolveArtifactTransitively( Artifact artifact )
    throws
        ArtifactResolutionException, ArtifactMetadataRetrievalException,
        ArtifactNotFoundException {
        
        resolveArtifact(artifact);

        ResolutionGroup rg =
            artifactMetadataSource.retrieve(
                artifact, localRepository, remoteRepositories );
        Set<Artifact> artifacts = rg.getArtifacts();
        
        ArtifactResolutionResult result =
            resolver.resolveTransitively(
                artifacts, artifact, remoteRepositories, localRepository,
                artifactMetadataSource );
        Set<Artifact> as = result.getArtifacts();
        
        List<Artifact> list = new ArrayList<Artifact>();
        list.add(artifact);
        list.addAll(as);

        return list;
    }

    
    // --------------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------------
    
    /**
     * Return a copy of the specified strings, with leading and trailing blank
     * characters (whitespace, tab and newline) omitted. If the specified list
     * is <code>null</code> or if its size is equals to 0, return a 0-length
     * array.
     */
    private static String[] trimHeadingAndTrailingBlanks( List<String> src ) {
        
        if( src==null || src.size()==0 ) {
            return new String[0];
        }
        
        String[] srcs = new String[src.size()];
        for (int i = 0; i < src.size(); i++) {
            String s = src.get(i);
            srcs[i] = Utils.trimHeadingAndTrailingBlanks(s);
        }
        return srcs;
    }
}
