/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Philippe Merle
 */

package org.objectweb.fractal.juliac;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DependencyResolutionRequiredException;
import org.apache.maven.artifact.metadata.ArtifactMetadataRetrievalException;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.model.Resource;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;

/**
 * This class manages the locations of the classpath which is created by {@link
 * AbstractJuliacMojo}.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 * @since 2.1
 */
public class JuliacMojoClassPathElements {
        
    public JuliacMojoClassPathElements( JuliacMojoItf jmi ) {
        this.jmi = jmi;
    }
    private JuliacMojoItf jmi;

    private List<Artifact> artifacts = new ArrayList<Artifact>();
    private List<String> classpaths = new ArrayList<String>();    

    
    // --------------------------------------------------------------------
    // Methods for retrieving elements from the classpath
    // --------------------------------------------------------------------
    
    public URL[] getURLs()
    throws
        MalformedURLException, ArtifactResolutionException,
        DependencyResolutionRequiredException,
        ArtifactMetadataRetrievalException, ArtifactNotFoundException {

        /*
         * Retrieve transitively all depending artifacts and add the
         * corresponding paths.
         */
        List<Artifact> artifacts =
            jmi.resolveArtifactTransitively(this.artifacts);
        for (Artifact artifact : artifacts) {
            jmi.resolveArtifact(artifact);
            File file = toFile(artifact);
            String path = file.getAbsolutePath();
            addPath(path);
        }
        
        /*
         * Discard the artifact list to avoid duplicating paths the next time
         * the URLs are requested.
         */
        artifacts.clear();
        
        /*
         * Retrieve the URLs corresponding to the paths.
         */
        URL[] urls = new URL[classpaths.size()];
        int i = 0;
        for (String classpath : classpaths) {
            File f = new File(classpath);
            URI uri = f.toURI();
            urls[i] = uri.toURL();
            i++;
        }
        
        return urls;
    }
    
    
    // --------------------------------------------------------------------
    // Methods for adding elements in the classpath
    // --------------------------------------------------------------------
    
    public void addResources( Collection<Resource> resources ) {
        for (Resource resource : resources) {
            String dir = resource.getDirectory();
            if( ! classpaths.contains(dir) ) {
                classpaths.add(dir);
            }
        }
    }

    public void addArtifacts( Collection<Artifact> artifacts ) {
        for (Artifact a : artifacts) {
            addArtifact(a);
        }
    }
    
    private void addArtifact( Artifact artifact ) {
        artifacts.add(artifact);
    }
    
    /**
     * Add the specified location descriptors to the classpath. A location
     * descriptor is either a file name or an artifact descriptor. The format of
     * an artifact descriptor is expected to be
     * <code>groudId:artifactId:version:classifier</code>.
     * 
     * @param projectBaseDir  the base directory storing the Maven project
     * @param locations       the array of location descriptors
     * @return                the paths corresponding to the locations
     * @throws MojoExecutionException  in case of bad artifact descriptor
     */
    public String[] addLocationDescriptors(
        File projectBaseDir, String[] locations )
    throws
        MojoExecutionException, ArtifactResolutionException,
        ArtifactNotFoundException, DependencyResolutionRequiredException {
        
        List<String> paths = new ArrayList<String>();
        
        for (String location : locations) {
            
            File f = new File(location);
            if( f.exists() ) {
                // Absolute path name or path name relative to the current dir
                String path = f.getAbsolutePath();
                addPath(path);
                paths.add(path);
            }
            else {
                f = new File(projectBaseDir,location);
                if( f.exists() ) {
                    // Path name relative to the Maven project base directory
                    String path = f.getAbsolutePath();
                    addPath(path);
                    paths.add(path);
                }                
                else {
                    /*
                     * Artifact.
                     */
                    String[] elts = location.split(":");
                    if( elts.length < 2 ) {
                        String msg =
                            "Expecting either a file name or an artifact descriptor of "+
                            "the form 'groudId:artifactId:version:classifier' (classifier and version can be omitted). "+
                            "Got '"+location+"'.";
                        throw new MojoExecutionException(msg);
                    }

                    String groupId = elts[0];
                    String artifactId = elts[1];
                    String version = elts.length>2 ? elts[2] : jmi.getPluginVersion();
                    String type = "jar";
                    String classifier = elts.length>3 ? elts[3] : null;

                    /*
                     * Resolve the artifact and its transitive dependencies and add
                     * them to the classpath.
                     */
                    Artifact a =
                        jmi.resolveArtifact(
                            groupId, artifactId, version, type, classifier );
                    addArtifact(a);
                    File afile = toFile(a);
                    String path = afile.getAbsolutePath();
                    paths.add(path);
                }
            }
        }
        
        return paths.toArray(new String[paths.size()]);
    }
    
    public void addOptimizationLevel( String ol )
    throws ArtifactResolutionException, ArtifactNotFoundException {
        
        String groupId = "org.objectweb.fractal.juliac";
        String artifactId = ol;
        String version = jmi.getPluginVersion();
        String type = "jar";
        String classifier = "";
        
        Artifact a =
            jmi.resolveArtifact(
                groupId, artifactId, version, type, classifier );
        addArtifact(a);
    }

    public void addPaths( Collection<String> paths ) {
        for (String elt : paths) {
            addPath(elt);
        }
    }

    public void addPath( String path ) {        
        if( classpaths.contains(path) ) {
            // Skip elements already inserted
            return;
        }
        classpaths.add(path);
    }
    

    // --------------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------------
    
    /**
     * Return the file corresponding to the specified artifact.
     * 
     * Most of the code in this method is copied from {@link
     * MavenProject#addArtifactPath(Artifact,List) which is not visible
     * (private).
     */
    private File toFile( Artifact a )
    throws DependencyResolutionRequiredException {
        
/*
  Removing the following code seems to have no impact for building Juliac and Tinfi.
  But when present, this introduces an issue with FraSCAti. Fixed by Philippe.
  
  The issue is related to the fact that both M/target/classes et M/target/M.jar
  being part of the classpath. FraSCAti scans the entire class path and merges
  each encountered .composite. Merging the same .composite twice introduces
  unwanted side effects.

        String groupId = a.getGroupId();
        String artifactId = a.getArtifactId();
        String version = a.getVersion();
        String refId = getProjectReferenceId( groupId, artifactId, version );
        Map<String,MavenProject> projectReferences = jmi.getProjectReferences();
        MavenProject project = projectReferences.get(refId);

        if ( project != null ) {
            Build build = project.getBuild();
            String path = build.getOutputDirectory();
            File file = new File(path);
            return file;
        }
        else {
*/
            File file = a.getFile();
            if( file == null ) {
                throw new DependencyResolutionRequiredException(a);
            }
            return file;
//        }
    }

    /*
     * Copied from {@link
     * MavenProject#getProjectReferenceId(String,String,String) which is not
     * visible (private static).
     */
/*
    private static String getProjectReferenceId(
        String groupId, String artifactId, String version ) {
        
        return groupId + ":" + artifactId + ":" + version;
    }
*/
}
